/*
 *
 * LICENSE AGREEMENT FOR Export to Proto.io PLUGIN
 *
 * Copyright (c) 2015, PROTOIO INC All rights reserved.

 * BSD-2-Clause License: http://opensource.org/licenses/BSD-2-Clause
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */



(function () {
    "use strict";
    var PLUGIN_ID = require("./package.json").name,
        MENU_ID = "protoio_all",
        MENU_LABEL = "$$$/JavaScripts/Generator/Proto.io/Menu=Proto.io Export All Artboards and Layers";
    
    var _generator = null,
        _config = null,
		_currentDocument=null;
		


    /*********** INIT ***********/
	//var photoshop=require("photoshop");
    var packageJSON = require('./package.json');
    var debug=packageJSON.debug;
    var async = require("async");
    var fs = require('fs');
    var yazl = require("yazl");
    var hiddenLayersInPage=[];
    var screens=[];
    var gVersion=packageJSON.versionInApp;
    var jsonOutput={"about":"Proto.io Photoshop generator export","version":gVersion,screens:[]};
    var gCoords={};
    var gCoordsHelper={};
    var gMode;
    var gExportHidden=true;
    var gGroupGroups=true;
    var gPSArtboards={};
    function init(generator, config) {
        _generator = generator;
        _config = config;

        console.log("initializing generator getting started tutorial with config %j", _config);
        _generator.addMenuItem("protoio_all", "$$$/JavaScripts/Generator/Proto.io All/Menu=Proto.io - Export All Artboards and Layers as Screens", true, false);
        _generator.addMenuItem("protoio_selected_artboards", "$$$/JavaScripts/Generator/Proto.io Selected/Menu=Proto.io - Export Selected Artboards as Screens", true, false);
        _generator.addMenuItem("protoio_selected_layers", "$$$/JavaScripts/Generator/Proto.io Assets/Menu=Proto.io - Export Selected Layers as Assets", true, false);

        _generator.onPhotoshopEvent("generatorMenuChanged", handleGeneratorMenuClicked);

    }

    /*********** EVENTS ***********/

    function handleGeneratorMenuClicked(event) {
        // Ignore changes to other menus
        var menu = event.generatorMenuChanged;
        if (!menu ) {
            return;
        }

       if(menu.name=="protoio_all") {
           gMode="all";
           requestEntireDocument();
       }else if(menu.name=="protoio_selected_artboards"){
           gMode="selected_artboards";
           requestEntireDocument();
       }else if(menu.name=="protoio_selected_layers"){
           gMode="selected_layers";
           requestEntireDocument();
       }

    }


    /*********** CALLS ***********/
    function requestEntireDocument(documentId) {
        if (!documentId) {
            console.log("Determining the current document ID");
        }
        _generator.getDocumentInfo(documentId).then(
            function (document) {
                populateDocumentPSArtboards(document);
            },
            function (err) {
                console.error("Error getting document info:", err);
            }
        ).done();
    }


    function exportDocument(Document){
        selectFolder(Document);
    }

    function doExportDocument(Document,folderSettings,selectedArtboards,selectedLayers){
        _currentDocument=Document;
        hiddenLayersInPage=[];
        screens=[];
        gCoords={};
        gCoordsHelper={};
        if(gMode=="selected_layers"){
            hiddenLayersInPage=selectedLayers;
            return bulkProcessInOrder(hiddenLayersInPage,folderSettings);
        }else{
            async.series([
               function(sCallback){
                   processArtboards(Document,selectedArtboards,sCallback ); //loop through artboards and generate struct of elements
                }], function(error,dummeData) {
                   return bulkProcessInOrder(hiddenLayersInPage,folderSettings);
                });


        }

    }


    function populateDocumentPSArtboards(document){
        gPSArtboards={};

        async.forEachSeries(document.layers,
            function (item,callback){
                isPSArtboardJSX(item,callback)
            }
            ,function(err){
                if(err){

                }else{
                    populateDocumentPSArtboardsDone(document);
                }

            })

    }
    function populateDocumentPSArtboardsDone(document){
        processDocument(document);
    }




    function processDocument(document){
        var message="All Artboards"
        var selectedArtboards=[];
        var selectedLayers=[];
        if(gMode=="selected_artboards"){
            message="Selected Arbtoards"
            selectedArtboards=validateArtboardsSelected(document);
            if(selectedArtboards.length==0){
                _generator.alert("You need to select one or more Artboards for export.\nIf you don't have any Artboards and you are selecting groups, then make sure they are located in the root and are prefixed with the '+' sign.")
                return;
            }
        }else if(gMode=="all"){
            selectedArtboards=getExportableArtboards(document);
            if(selectedArtboards.length==0){
                _generator.alert("No Artboards found.\nYou may either create Artboards or prefix one or more root Groups with the '+' sign.")
                return;
            }
        }else if(gMode=="selected_layers"){
            //TODO: implement this today
            //make sure some layers are selected.
            //loop through all layers and add the ones selected to hiddenLayersInPage
            //call bulkProcess directly without calling processArtboards
            selectedLayers=validateLayersSelected(document,document);
            if(selectedLayers.length==0){
                _generator.alert("You need to select one or more layers.")
                return;
            }

        }

        openDialogForExport(document, selectedArtboards, selectedLayers,"");
        return;



       /* async.forEachSeries(selectedArtboards, function (selectedArtboard,eCallback){

                async.waterfall([

                        function(wCallback) {
                            getArtboardBackgroundColor(selectedArtboard, wCallback);
                        }],
                    function (error, background) {

                        if (background !== "rgba(0,0,0,0)") {
                            eCallback(new Error('No transparent background'),null);
                        }
                        eCallback(error,null);
                    });

            },
            function(err){

                if (err) {
                    openDialogForNonTransparent(document, selectedArtboards, selectedLayers, message);
                } else {
                    openDialogForExport(document, selectedArtboards, selectedLayers, message);
                }

            });*/



    }

    function openDialogForNonTransparent(document, selectedArtboards, selectedLayers, message){


        var jsxPluginFile=__dirname+"/protoio-no-transparent-modal.jsx";
        _generator.evaluateJSXFile(jsxPluginFile, { dummy: ''}).then(function(result){
                if(result!="Cancel"){
                    openDialogForExport(document, selectedArtboards, selectedLayers, message);
                }
            },
            function(err){
                console.error("Error:",err);
                _generator.alert(err);

            }

        );

    }

    function openDialogForExport(document, selectedArtboards, selectedLayers, message){


        var jsxPluginFile=__dirname+"/protoio.jsx";
        _generator.evaluateJSXFile(jsxPluginFile,{exportHidden:gExportHidden,groupGroups:gGroupGroups,message:message,version:gVersion}).then(function(result){
                //_generator.alert("got result"+result);
                if(result!="Cancel"){
                    gExportHidden=result.exportHidden;
                    gGroupGroups=result.groupGroups;
                    selectFolder(document,selectedArtboards,selectedLayers);
                }

            },
            function(err){
                console.error("Error:",err);
                _generator.alert(err);

            }

        );


    }

    function selectFolder(Document,selectedArtboards,selectedLayers){
        _generator.evaluateJSXString('var outFolder = Folder.selectDialog("Select a folder to save export file");outFolder.fsName').then(function(result) {
            if(!result) {
                return false;
            }
            var tempFolder=result;
            _generator.alert('Export process started.\nYou will be notified when the process completes.');
            var folderSettings=initFolders(Document,tempFolder);
            doExportDocument(Document,folderSettings,selectedArtboards,selectedLayers)
        });
    }


    function initFolders(Document,tempFolder){
        var _document=Document;
        var dirsep="/";
        var isWin = /^win/.test(process.platform);
        if (isWin)
            dirsep="\\";
        var filename = _document.file;
        var docName = filename.split(dirsep).slice(-1)[0].replace('.psd','');


        //var sourceFolder="/Users/alexisodysseos/Documents/Proto-Photoshop/";
        var sourceFolder=tempFolder;
        var fileName=sourceFolder+"Inbox/"+docName+".psd";
        //var outFolder=sourceFolder+"Exports/"+docName+".temp/";
        var outFolder=sourceFolder+"/"+docName+".protoiotemp/";

        if(gMode=="selected_layers"){
            outFolder=sourceFolder+"/";
        }else{
            var dir = outFolder;
            deleteFolderRecursive(dir);
            if (!fs.existsSync(dir)){
                fs.mkdirSync(dir);
            }

        }


        return {"outFolder":outFolder,"docName":docName,"folderToSandbox":sourceFolder+dirsep}

    }

    function processArtboards(Document,selectedArtboards, callback){
        var _document=Document;
        var artboardsToExport=selectedArtboards;
        if(!artboardsToExport || artboardsToExport.length==0) {
            artboardsToExport = getExportableArtboards(_document);
        }

        async.forEachSeries(artboardsToExport, function (currentArtboard,eCallback){
                // var currentArtboard=artboardsToExport[i];

                var coords=getCoordsFromBounds(currentArtboard.bounds||currentArtboard.boundsWithFX);
                var screenInfo={"title":currentArtboard.name,"id":currentArtboard.id,size:{"width":coords.width,"height":coords.height},"position":{"x":coords.x,"y":coords.y}}


                async.waterfall([

                    function(wCallback) {
                        getArtboardBackgroundColor(currentArtboard, wCallback);
                    }],
                    function (error, background) {

                        if (background != null) {
                            currentArtboard.backgroundColor = background;
                            screenInfo.backgroundColor = background;
                        }

                        var items=processExportableChildren(currentArtboard,currentArtboard.layers,"","");
                        screenInfo.items=items;
                        screens.push(screenInfo);
                        eCallback(error,null);
                    });


            }
            ,function(err){

                jsonOutput.screens=screens;
                callback(null,null);

            });

        // for (var i=0;i<artboardsToExport.length;i++){
        //     var currentArtboard=artboardsToExport[i];
        //     var coords=getCoordsFromBounds(currentArtboard.bounds||currentArtboard.boundsWithFX);
        //     var screenInfo={"title":currentArtboard.name,"id":currentArtboard.id,size:{"width":coords.width,"height":coords.height},"position":{"x":coords.x,"y":coords.y}}
        //     var items=processExportableChildren(currentArtboard,currentArtboard.layers,"","");
        //     screenInfo.items=items;
        //     screens.push(screenInfo);
        // }
        //
        // jsonOutput.screens=screens;
    }

    function getExportableArtboards(Document){
        var _document=Document;
        var _layers=Document.layers;
        var layersCount=_layers.length;
        if(layersCount==0){return [];}
        var rValue=[];
        var currentArtboard;
        for (var i=0;i<layersCount;i++) {
            currentArtboard=_layers[i];
            if(isArtboardExportable(currentArtboard)){
                rValue.push(currentArtboard);
            }
        }
        return rValue;

    }
    function validateArtboardsSelected(Document){
        if(!(Document.selection && Document.selection.length>0)){
            return [];
        }
        var selection=Document.selection;
        var artboardsToExport=[];
        for (var i=0;i<selection.length;i++){
            var currentIndex=selection[i];
            for (var n=0;n<Document.layers.length;n++){
                var currentArtboard=Document.layers[n];
                if(currentArtboard.index==currentIndex && isArtboardExportable(currentArtboard)){
                    artboardsToExport.push(currentArtboard);
                }
            }
        }
        return artboardsToExport;
    }

    function validateLayersSelected(Document,layer){
        if(!(Document.selection && Document.selection.length>0)){
            return [];
        }
        var selection=Document.selection;
        var rValue=[];
        if(isLayerSelected(Document,layer) && layer.type!="layerSection"){
            rValue.push(layer);
        }
        if(layer.layers && layer.layers.length>0){
            for( var i=0;i<layer.layers.length;i++){
                var currentLayer=layer.layers[i];
                var selectedSubLayers= validateLayersSelected(Document,currentLayer);
                if(selectedSubLayers.length>0){
                    for(var n=0;n<selectedSubLayers.length;n++){
                        rValue.push(selectedSubLayers[n]);
                    }
                }
            }
        }
        return rValue;
    }

    function isLayerSelected(Document,layer){
        if(!(Document.selection && Document.selection.length>0)){
            return false;
        }
        for(var i=0;i<Document.selection.length;i++){
            if(layer.index==Document.selection[i]){
                return true;
            }
        }
        return false;

    }
    function isArtboardExportable(Artboard){
        if(gPSArtboards[Artboard.id]){return true;}
        var artboardName=Artboard.name;
        return artboardName.substr(0,1)=="+" && Artboard.type=="layerSection";
    }

    function isPSArtboardJSX(Artboard,callback){
        var LayerId=Artboard.id;
        var JSXScript="\
            var ref = new ActionReference();\
            ref.putIdentifier(charIDToTypeID('Lyr '), "+LayerId+"  );\
            var isArtboardLayer =false;\
            try{\
                isArtboardLayer = executeActionGet(ref).getBoolean(stringIDToTypeID('artboardEnabled'));\
            }catch(e){\
            };\
            if(isArtboardLayer){\
            String('"+LayerId+"');\
            }else{\
            String('');\
            }\
        "
        _generator.evaluateJSXString(JSXScript).then(
            function(result){
                if(result!=""){
                    console.log(result);
                    gPSArtboards[result]=true;
                }

                callback();

            },
            function(err){
                console.log(err);
                callback();


            });


       // return isArtboardLayer;

    }

    function getArtboardBackgroundColor(Artboard,callback){
        var LayerId=Artboard.id;
        var JSXScript="var classLayer = charIDToTypeID('Lyr ');\n" +
            "var idnull = stringIDToTypeID( \"null\" );\n" +
            "\n" +
            "var classProperty = charIDToTypeID('Prpr');\n" +
            "var typeOrdinal = charIDToTypeID('Ordn');\n" +
            "var enumTarget = charIDToTypeID('Trgt'); \n" +
            "\n" +
            "var idred = charIDToTypeID( 'Rd  ' );\n" +
            "var idgreen = charIDToTypeID( 'Grn ' );\n" +
            "var idblue = charIDToTypeID( 'Bl  ' );\n" +
            "var actionGet = charIDToTypeID( \"getd\" );\n" +
            "\n" +
            "var idtop = stringIDToTypeID( \"top\" );\n" +
            "var idleft = stringIDToTypeID( \"left\" );\n" +
            "var idright = stringIDToTypeID( \"right\" );\n" +
            "var idbottom = stringIDToTypeID( \"bottom\" );\n" +
            "\n" +
            "var idartboard = stringIDToTypeID( \"artboard\" );\n" +
            "var idartboardRect = stringIDToTypeID( \"artboardRect\" );\n" +
            "var idartboardColor = stringIDToTypeID( \"color\" );\n" +
            "\n" +
            "function TargetLayerRef(inProp) {\n" +
            "\tvar theRef = new ActionReference();\n" +
            "    theRef.putIdentifier(charIDToTypeID('Lyr '), "+ LayerId + "  );\n" +
            "\t//theRef.putProperty(classProperty, inProp);\n" +
            "\t//theRef.putEnumerated(classLayer, typeOrdinal, enumTarget);\n" +
            "\n" +
            "\treturn theRef;\n" +
            "}\n" +
            "\n" +
            "function RefArtboardRect(theRef) {\n" +
            "\tvar getDescriptor = new ActionDescriptor();\n" +
            "\tgetDescriptor.putReference(idnull,theRef);\n" +
            "\tvar artboardRect = executeAction( actionGet, getDescriptor, DialogModes.NO )\n" +
            "\t\t\t\t\t\t.getObjectValue(idartboard).getObjectValue(idartboardRect);\n" +
            "\n" +
            "\treturn {\n" +
            "\t\ttop: artboardRect.getInteger(idtop), \n" +
            "\t\tleft: artboardRect.getInteger(idleft),\n" +
            "\t\tbottom: artboardRect.getInteger(idbottom),\n" +
            "\t\tright: artboardRect.getInteger(idright)\n" +
            "\t};\n" +
            "}\n" +
            "\n" +
            "function TargetArtboardRect() {\n" +
            "\treturn RefArtboardRect(TargetLayerRef(idartboard));\n" +
            "}\n" +
            "\n" +
            "function RefArtboardColor(theRef) {\n" +
            "\tvar getDescriptor = new ActionDescriptor();\n" +
            "\tgetDescriptor.putReference(idnull,theRef);\n" +
            "\n" +
            "\tvar artboardColor = executeAction( actionGet, getDescriptor, DialogModes.NO )\n" +
            "\t\t\t.getObjectValue(idartboard).getObjectValue(idartboardColor);\n" +
            "            \n" +
            "           \n" +
            "    var idartboardBgType = stringIDToTypeID( \"artboardBackgroundType\" );\n" +
            "    var artboard = executeAction( actionGet, getDescriptor, DialogModes.NO).getObjectValue(idartboard);\n" +
            "    var bgType = artboard.getInteger(idartboardBgType);\n" +
            "     \n" +
            "\n" +
            "    var returnValue = \"\";\n" +
            "    if (bgType == 1) { // white background\n" +
            "        \n" +
            "        returnValue = \"rgba(255,255,255,1.00)\";\n" +
            "        \n" +
            "    } else if (bgType == 2) { // black background\n" +
            "        \n" +
            "        returnValue = \"rgba(0,0,0,1.00)\";\n" +
            "    \n" +
            "    } else if \n" +
            "(bgType == 3) { //transparent background\n" +
            "        \n" +
            "        returnValue = \"rgba(0,0,0,0)\";\n" +
            "        \n" +
            "    } else{\n" +
            "        returnValue = \"rgba( \" + artboardColor.getInteger(idred) + \", \" + artboardColor.getInteger(idgreen) + \",\" + artboardColor.getInteger(idblue) + \",1.00)\";\n" +
            "    }\n" +
            "    //rgba(0,0,0,1.00)\n" +
            "    //rgba(artboardColor.getInteger(idred),artboardColor.getInteger(idgreen),artboardColor.getInteger(idblue),1.00)\n" +
            "        \n" +
            "\treturn returnValue;\n" +
            "}\n" +
            "\n" +
            "function TargetArtboardColor() {\n" +
            "\treturn RefArtboardColor(TargetLayerRef(idartboard));\n" +
            "}\n" +
            "\n" +
            "// =======================================================\n" +
            "\n" +
            "var propArtboardEnabled = stringIDToTypeID( \"artboardEnabled\" );\n" +
            "\n" +
            "function TargetIsArtboard() {\n" +
            "\tvar getDescriptor = new ActionDescriptor();\n" +
            "\tgetDescriptor.putReference(idnull,TargetLayerRef(propArtboardEnabled));\n" +
            "\treturn executeAction( actionGet, getDescriptor, DialogModes.NO ).getBoolean(propArtboardEnabled);\n" +
            "}\n" +
            "\n" +
            "function RectToString(inRect) {\n" +
            "\treturn \"\" + inRect.top + \",\"  + inRect.left + \",\"  + inRect.bottom + \",\" + inRect.right;\n" +
            "\t}\n" +
            "\n" +
            "// =======================================================\n" +
            "\n" +
            "function RGBToString(inRGB) {\n" +
            "\treturn \"\" + inRGB.red + \",\"  + inRGB.green + \",\"  + inRGB.blue;\n" +
            "\t}\n" +
            "\n" +
            "\n" +
            "// =======================================================\n" +
            "// Display artboard rectangle of selected artboard.\n" +
            "\n" +
            "function DisplayTargetArtboard() {\n" +
            "\tif (TargetIsArtboard()) {\n" +
            "\t\tvar artboardRect = TargetArtboardRect(),\n" +
            "\t\t\tartboardColor = TargetArtboardColor();\n" +
            "\t\t//return String(\"Artboard Rect: \" + RectToString(artboardRect) + \"\\nBackground Color: \" + RGBToString(artboardColor) );\n" +
            "        return String(artboardColor);\n" +
            "\t}\n" +
            "}\n" +
            "\n" +
            "// =======================================================\n" +
            "\n" +
            "DisplayTargetArtboard();";


        _generator.evaluateJSXString(JSXScript).then(
            function(result){
                console.log(result);

                if (result == "") {
                    result = null;
                }

                callback(null, result);

            },
            function(err){
                console.log(err);
                callback(err, null);
            });

    }

    function processExportableChildren(currentArtboard,Layers,pathId,pathName){
        var _layers=Layers;

        if (typeof _layers === 'undefined') { return;}

        var layersCount=_layers.length;
        var items=[];
        var itemInfo={};
        if(layersCount==0){return;}
        for (var i=layersCount-1;i>=0;i--) {
            var layer=_layers[i];
            var isExportable=isLayerExportable(layer);
            var newPathId=pathId+"/"+layer.id;
            var newPathName=pathName+"/"+layer.name;
            itemInfo={"id":layer.id,"name":layer.name,"path":pathName,"pathID":pathId,"hidden":!layer.visible}
            var exportingGroupGroup=false;

            if(layer.layers){
                if(gGroupGroups && isGroupGroup(layer)){
                    exportingGroupGroup=true;
                }
                if(isExportable && !exportingGroupGroup){
                    //this is a group
                    itemInfo.isGroup=1;
                    items.push(itemInfo);
                    var childItems=processExportableChildren(currentArtboard,layer.layers,newPathId,newPathName);
                    for(var n=0;n<childItems.length;n++){
                        var currentItem=childItems[n];
                        items.push(currentItem);
                    }
                }
            }
            if (isExportable){
                var coords=getCoordsFromBounds(layer.bounds);
                if(layer.boundsWithFX){
                    coords=getCoordsFromBounds(layer.boundsWithFX);
                }
                var artCoords=getCoordsFromBounds(currentArtboard.bounds);
                var positionX=-artCoords.x+coords.x;
                var positionY=-artCoords.y+coords.y;

                // if (currentArtboard.hasOwnProperty('backgroundColor') && currentArtboard.backgroundColor !== "rgba(0,0,0,0)") {
                //     getActualBoundOfAnObject(layer.id, artCoords, coords);
                // }


                // positionX = (positionX >= 0) ? positionX : 0;
                // positionY = (positionY >= 0) ? positionY : 0;

                itemInfo.fileName=fixFileName(layer.name)+"~"+layer.id+".png";
                itemInfo.size={"width":coords.width,"height":coords.height};
                itemInfo.position={"x":positionX,"y":positionY};
                if(!layer.layers || ( layer.layers && exportingGroupGroup)){
                    //we already added the group
                    items.push(itemInfo);
                }

                hiddenLayersInPage.push(layer);
            }
        }
        return items;

    }

    function getActualBoundOfAnObject(layerId, artCoords, coords){

        var positionX=-artCoords.x+coords.x;
        var positionY=-artCoords.y+coords.y;


        var x = (positionX >= 0) ? positionX : 0;
        var y = (positionY >= 0) ? positionY : 0;

        var width = 0;

        if ( positionX < 0 ) {
            width = coords.width + positionX;

            if ( width >  artCoords.width ) {
                width = width - (width - artCoords.width);
            }

        } else {
            width = coords.width;

            if ( (width + positionX) >  artCoords.width ){
                width = width - ( (x + width) - artCoords.width );
            }


        }

        var height = 0;

        if ( positionY < 0 ) {
            height = coords.height + positionY;

            if ( height >  artCoords.height ) {
                height = height - (height - artCoords.height);
            }

        } else {
            height = coords.height;

            if ( (height + positionY) >  artCoords.height ){
                height = height - ( (y + height) - artCoords.height );
            }

        }



        gCoordsHelper[layerId] = {
            x : x,
            y : y,
            width : width,
            height : height

        };


    }

    function bulkProcessInOrder(HiddenLayersInPage,folderSettings){
        //export and hide layer in order
        async.forEachSeries(HiddenLayersInPage,
            function (item,callback){
                //console.log("saving layer:"+item.name);
                if((item.layers || item.type=="layerSection") && (!gGroupGroups || (gGroupGroups && !isGroupGroup(item)))){
                    callback()
                }else{
                    saveLayerBitmap(_currentDocument.id,item.id,fixFileName(item.name)+"~"+item.id+".png",folderSettings,callback);
                }
            }
            ,function(err){
                exportWholeDocument(folderSettings);
                console.log("export complete");
            })
    }

    function exportWholeDocument(folderSettings){
        console.log("saving screen");
        //20151021 saveDocumentBitMap(_currentDocument.id,"Screen-"+docName+".png",folderSettings,cleanupLayers);
        cleanupLayers();
        updateLayerSizes();
        if(gMode=="selected_layers"){
            moveFilesInPlace(folderSettings);
        }else{
            writeJSON(folderSettings);
            createArchive(folderSettings);
        }
    }

    function cleanupLayers(){
        var HiddenLayersInPage=hiddenLayersInPage;
        //Show all layers that were hidden
        async.forEachSeries(HiddenLayersInPage,
            function (item,callback) {
                //20151021 setLayerVisibility(item.id, true);
                callback();
            }
            ,function(err){
                //all done
                console.log("cleanup complete");

            })
    }

    function updateLayerSizes(){

        var arrayOfScreensRemoval = [];

        for (var i=0;i<screens.length;i++){

            if (typeof screens[i].items == 'undefined'){
                arrayOfScreensRemoval.push(i);
                continue;
            }

            for (var n=0;n<screens[i].items.length;n++){
                var currentItem=screens[i].items[n];
                var coords=gCoords[currentItem.id];
                if(currentItem.position && coords){
                    //currentItem.position.x=coords.x;
                    //currentItem.position.y=coords.y;
                    currentItem.size.width=coords.width;
                    currentItem.size.height=coords.height;
                }
            }
        }

        if (arrayOfScreensRemoval.length > 0) {
            for (var i = 0; i < arrayOfScreensRemoval.length; i++) {
                screens.splice(arrayOfScreensRemoval[i], 1);
            }
        }

    }

    function writeJSON(folderSettings){
        var infoFile=folderSettings.outFolder+"/info.json";
        fs.writeFileSync(infoFile, JSON.stringify(jsonOutput));

    }

    function createArchive(folderSettings){
        var outFolder=folderSettings.outFolder;
        var folderToSandbox=folderSettings.folderToSandbox;
        var dir=fs.readdirSync(outFolder);
        var zipfile = new yazl.ZipFile();
        for(var i=0;i<dir.length;i++){
            zipfile.addFile(outFolder+dir[i], dir[i]);
        }
        var outFile=folderToSandbox+folderSettings.docName+".protoio";
        zipfile.outputStream.pipe(fs.createWriteStream(outFile)).on("close", function(a) {
            if(!debug){
                deleteFolderRecursive(outFolder);
            }
            _generator.alert("Proto.io package saved here:\n"+outFile);
        });
        zipfile.end();

    }

    function moveFilesInPlace(folderSettings){
        _generator.alert("Assets exported here:\n"+folderSettings.folderToSandbox);
    }

    function exportLayer(Layer){
        var layer=Layer;
        console.log("exporting"+layer.name);
    }

    function isLayerExportable(Layer){
        var _layer=Layer;
        return (!_layer.clipped && (gExportHidden || !gExportHidden && _layer.visible));
        var i=_layer.name.indexOf(".png")
        return i!=-1;

    }

    function isGroupGroup(Layer){
        return Layer.name.substr(0,1)=="@";
    }

    function saveLayerBitmap(DocumentId,LayerId,SaveFileName,folderSettings,callback){


        var settings = {};
        //settings.includeAncestorMasks = true;



        _generator.getPixmap(DocumentId,LayerId,settings).then(
            function(pixmap){
                //console.log("got Pixmap for layerID "+LayerId+":wxh="+pixmap.width+" x "+pixmap.height);
                gCoords[LayerId]={"x":pixmap.bounds.top,"y":pixmap.bounds.left,"width":pixmap.width,"height":pixmap.height};
                // gCoordsHelper[LayerId];
                var pixmapSettings = {format:"png",ppi:72,background:[0,0,0,0.0]};

                // if (gCoordsHelper.hasOwnProperty(LayerId)) {
                //     // pixmapSettings.extract = gCoordsHelper[LayerId];
                //     // pixmapSettings.extract.x = 0;
                //     // pixmapSettings.extract.y = 0;
                //     // gCoords[LayerId]= gCoordsHelper[LayerId];
                // }

                _generator.savePixmap(pixmap,folderSettings.outFolder+ SaveFileName,pixmapSettings).then(function(){
                //_generator.savePixmap(pixmap,folderSettings.outFolder+ SaveFileName,{format:"png",ppi:72}).then(function(){
                    callback();
                },function(err){
                    callback();
                });
            },
            function(err){
                console.error("err pixmap:",err);
            }).done();

    }

    function saveDocumentBitMap(DocumentId,SaveFileName,folderSettings,callback){
        _generator.getDocumentPixmap(DocumentId,{}).then(
            function(pixmap){
                _generator.savePixmap(pixmap,folderSettings.outFolder+ SaveFileName,{format:"jpg",quality:100,ppi:72});
                callback();
            },
            function(err){
                console.error("err pixmap:",err);
            }).done();

    }

    function setLayerVisibility(LayerId,Visible){
        if(Visible){
            var changeVisibilityString = " var ref = new ActionReference(); \
                                  ref.putIdentifier(charIDToTypeID('Lyr '), " + LayerId + " ); \
                                  var desc = new ActionDescriptor(); \
                                  desc.putReference(charIDToTypeID('null'), ref); \
                                  desc.putBoolean(charIDToTypeID('MkVs'), false); \
                                  executeAction( charIDToTypeID('Shw '), desc);"
        }else{

            var changeVisibilityString = " var ref = new ActionReference(); \
                                      ref.putIdentifier(charIDToTypeID('Lyr '), " + LayerId + " ); \
                                      var desc = new ActionDescriptor(); \
                                      desc.putReference(charIDToTypeID('null'), ref); \
                                      desc.putBoolean(charIDToTypeID('MkVs'), false); \
                                      executeAction( charIDToTypeID('Hd  '), desc);"
        }
        _generator.evaluateJSXString(changeVisibilityString);

    }

	function getFirstLayerBitmapAndSaveWithOptions(PassDocument){
        //SEQUENCE 5
	    var _document = PassDocument;
	    console.log(_document.id,_document.layers[0].id);
	    _generator.getPixmap(_document.id,_document.layers[0].id,{}).then(
	    function(pixmap){
	        console.log("got Pixmap: "+pixmap.width+" x "+pixmap.height);
	        _generator.savePixmap(pixmap,outPath+ 'out.png',{format:"jpg",quality:100,ppi:72});
	    },
	    function(err){
	        console.error("err pixmap:",err);
	    }).done();
	}

    function updateMenuState(enabled) {
        console.log("Setting menu state to", enabled);
        _generator.toggleMenu(MENU_ID, true, enabled);
    }

    
/*********** ALEXISO ***********/
	function openDocument_jsx(fileName){
        //runs /Library/Application Support/Adobe/Startup Scripts CC/Adobe Photoshop/openDocument.jsx
		return openDocument(fileName)
	}

	function getDocumentHeight_jsx(callback){
		//var _callback=callback;
		photoshop.invoke(callDocumentHeight_jsx, function(error, result){
			if(error){
				console.log(error)		
			}else{
				console.log("calling callback with "+result);
                console.log(callback);
                callback(result);
                //gotDocumentHeight(result);    
			}
		});
	}

	function callDocumentHeight_jsx(){
		return getDocumentHeight();
	}


    function getCoordsFromBounds(lb){
        var height = lb.bottom - lb.top;
        var width = lb.right - lb.left ;

        var coords = {
            x: lb.left,
            y: lb.top,
            height: height,
            width: width
        }

        return coords;
    }
	/*********** HELPERS ***********/
    function fixFileName(fileName){
        return fileName.replace(/[^a-z0-9@]/gi, '_').toLowerCase();
    }


    function sendJavascript(str){
        _generator.evaluateJSXString(str).then(
            function(result){
                console.log(result);
            },
            function(err){
                console.log(err);
            });
    }


    function stringify(object) {
        try {
            return JSON.stringify(object, null, "    ");
        } catch (e) {
            console.error(e);
        }
        return String(object);
    }

    function deleteFolderRecursive(path) {
        if( fs.existsSync(path) ) {
            fs.readdirSync(path).forEach(function(file,index){
                var curPath = path + "/" + file;
                if(fs.lstatSync(curPath).isDirectory()) { // recurse
                    deleteFolderRecursive(curPath);
                } else { // delete file
                    fs.unlinkSync(curPath);
                }
            });
            fs.rmdirSync(path);
        }
    };

    exports.init = init;

    // Unit test function exports
    exports._setConfig = function (config) { _config = config; };
    
}());